@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2005 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,23 +31,26 @@
 #include <vector>
 
 #include <sys/types.h>
-#include <dirent.h>
-#include <sys/stat.h>
-#include <unistd.h>
+#include "safeunistd.h"
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
-#include <fcntl.h>
-#include <errno.h>
+#include "safefcntl.h"
+#include "safeerrno.h"
+#include <time.h>
 
 #include <xapian.h>
 
 #include "commonhelp.h"
+#include "diritor.h"
 #include "hashterm.h"
-#include "indextext.h"
 #include "loadfile.h"
 #include "md5wrap.h"
 #include "metaxmlparse.h"
 #include "myhtmlparse.h"
+#include "sample.h"
+#include "stringutils.h"
+#include "utf8convert.h"
 #include "utils.h"
 #include "values.h"
 #include "xmlparse.h"
@@ -72,8 +75,15 @@
 # define pclose _pclose
 #endif
 
+#ifndef HAVE_MKDTEMP
+extern char * mkdtemp(char *);
+#endif
+
 using namespace std;
 
+#define TITLE_SIZE 128
+#define SAMPLE_SIZE 512
+
 #define PROG_NAME "omindex"
 #define PROG_DESC "Index static website data via the filesystem"
 
@@ -85,7 +95,9 @@ static string indexroot;
 static string baseurl;
 static Xapian::WritableDatabase db;
 static Xapian::Stem stemmer("english");
+static Xapian::TermGenerator indexer;
 static vector<bool> updated;
+static string tmpdir;
 
 inline static bool
 p_notalnum(unsigned int c)
@@ -93,38 +105,6 @@ p_notalnum(unsigned int c)
     return !isalnum(static_cast<unsigned char>(c));
 }
 
-/* Truncate a string to a given maxlength, avoiding cutting off midword
- * if reasonably possible. */
-string
-truncate_to_word(string & input, string::size_type maxlen)
-{
-    string output;
-    if (input.length() <= maxlen) {
-	output = input;
-    } else {
-	output = input.substr(0, maxlen);
-
-	string::size_type space = output.find_last_of(WHITESPACE);
-	if (space != string::npos && space > maxlen / 2) {
-	    string::size_type nonspace;
-	    nonspace = output.find_last_not_of(WHITESPACE, space);
-	    if (nonspace != string::npos) output.erase(nonspace);
-	}
-
-	if (output.length() == maxlen &&
-	    !isspace(static_cast<unsigned char>(input[maxlen]))) {
-	    output += "...";
-	} else {
-	    output += " ...";
-	}
-    }
-
-    // replace newlines with spaces
-    size_t i = 0;
-    while ((i = output.find('\n', i)) != string::npos) output[i] = ' ';
-    return output;
-}
-
 static string
 shell_protect(const string & file)
 {
@@ -148,6 +128,23 @@ shell_protect(const string & file)
     return safefile;
 }
 
+static bool ensure_tmpdir() {
+    if (!tmpdir.empty()) return true;
+
+    const char * p = getenv("TMPDIR");
+    if (!p) p = "/tmp";
+    char * dir_template = new char[strlen(p) + 15 + 1];
+    strcpy(dir_template, p);
+    strcat(dir_template, "/omindex-XXXXXX");
+    p = mkdtemp(dir_template);
+    if (p) {
+	tmpdir.assign(dir_template);
+	tmpdir += '/';
+    }
+    delete dir_template;
+    return (p != NULL);
+}
+
 struct ReadError {};
 struct NoSuchFilter {};
 
@@ -184,6 +181,48 @@ stdout_to_string(const string &cmd)
     return out;
 }
 
+static void get_pdf_metainfo(const string & safefile, string &title, string &keywords) {
+    try {
+	string pdfinfo = stdout_to_string("pdfinfo -enc UTF-8 " + safefile);
+
+	string::size_type idx;
+
+	if (strncmp(pdfinfo.c_str(), "Title:", 6) == 0) {
+	    idx = 0;
+	} else {
+	    idx = pdfinfo.find("\nTitle:");
+	}
+	if (idx != string::npos) {
+	    if (idx) ++idx;
+	    idx = pdfinfo.find_first_not_of(' ', idx + 6);
+	    string::size_type end = pdfinfo.find('\n', idx);
+	    if (end != string::npos) {
+		if (pdfinfo[end - 1] == '\r') --end;
+		end -= idx;
+	    }
+	    title = pdfinfo.substr(idx, end);
+	}
+
+	if (strncmp(pdfinfo.c_str(), "Keywords:", 9) == 0) {
+	    idx = 0;
+	} else {
+	    idx = pdfinfo.find("\nKeywords:");
+	}
+	if (idx != string::npos) {
+	    if (idx) ++idx;
+	    idx = pdfinfo.find_first_not_of(' ', idx + 9);
+	    string::size_type end = pdfinfo.find('\n', idx);
+	    if (end != string::npos) {
+		if (pdfinfo[end - 1] == '\r') --end;
+		end -= idx;
+	    }
+	    keywords = pdfinfo.substr(idx, end);
+	}
+    } catch (ReadError) {
+	// It's probably best to index the document even if pdfinfo fails.
+    }
+}
+
 static void
 index_file(const string &url, const string &mimetype, time_t last_mod, off_t size)
 {
@@ -231,6 +270,8 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	md5_string(text, md5);
     } else if (mimetype == "text/plain") {
 	try {
+	    // Currently we assume that text files are UTF-8.
+	    // FIXME: What charset is the file?  Look for BOM?  Look at contents?
 	    dump = file_to_string(file);
 	} catch (ReadError) {
 	    cout << "can't read \"" << file << "\" - skipping\n";
@@ -239,72 +280,57 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	md5_string(dump, md5);
     } else if (mimetype == "application/pdf") {
 	string safefile = shell_protect(file);
-	string cmd = "pdftotext " + safefile + " -";
-	//string cmd = "pdftotext -enc UTF-8 " + safefile + " -";
+	string cmd = "pdftotext -enc UTF-8 " + safefile + " -";
 	try {
 	    dump = stdout_to_string(cmd);
 	} catch (ReadError) {
 	    cout << "\"" << cmd << "\" failed - skipping\n";
 	    return;
 	}
-
-	try {
-	    string pdfinfo = stdout_to_string("pdfinfo " + safefile);
-	    //string pdfinfo = stdout_to_string("pdfinfo -enc UTF-8 " + safefile);
-
-	    string::size_type idx;
-
-	    if (strncmp(pdfinfo.c_str(), "Title:", 6) == 0) {
-		idx = 0;
-	    } else {
-		idx = pdfinfo.find("\nTitle:");
-	    }
-	    if (idx != string::npos) {
-		if (idx) ++idx;
-		idx = pdfinfo.find_first_not_of(' ', idx + 6);
-		string::size_type end = pdfinfo.find('\n', idx);
-		if (end != string::npos) {
-		    if (pdfinfo[end - 1] == '\r') --end;
-		    end -= idx;
-		}
-		title = pdfinfo.substr(idx, end);
-	    }
-
-	    if (strncmp(pdfinfo.c_str(), "Keywords:", 9) == 0) {
-		idx = 0;
-	    } else {
-		idx = pdfinfo.find("\nKeywords:");
-	    }
-	    if (idx != string::npos) {
-		if (idx) ++idx;
-		idx = pdfinfo.find_first_not_of(' ', idx + 9);
-		string::size_type end = pdfinfo.find('\n', idx);
-		if (end != string::npos) {
-		    if (pdfinfo[end - 1] == '\r') --end;
-		    end -= idx;
-		}
-		keywords = pdfinfo.substr(idx, end);
-	    }
-	} catch (ReadError) {
-	    // It's probably best to index the document even if pdfinfo fails.
-	}
+	get_pdf_metainfo(safefile, title, keywords);
     } else if (mimetype == "application/postscript") {
-	string cmd = "pstotext " + shell_protect(file);
-	try {
-	    dump = stdout_to_string(cmd);
-	} catch (ReadError) {
-	    cout << "\"" << cmd << "\" failed - skipping\n";
+	// There simply doesn't seem to be a Unicode capabable PostScript to
+	// text convertor (e.g. pstotext always outputs ISO-8859-1).  The only
+	// solution seems to be to convert via PDF using ps2pdf and then
+	// pdftotext.  This gives plausible looking UTF-8 output for some
+	// Chinese PostScript files I found using Google.  It also has the
+	// benefit of allowing us to extract meta information from PostScript
+	// files.
+	if (!ensure_tmpdir()) {
+	    // FIXME: should this be fatal?  Or disable indexing postscript?
+	    cout << "Couldn't create temporary directory (" << strerror(errno) << ") - skipping" << endl;
 	    return;
 	}
-    } else if (mimetype.substr(0, 24) == "application/vnd.sun.xml." ||
-	       mimetype.substr(0, 35) == "application/vnd.oasis.opendocument.")
+	string tmpfile = tmpdir + "/tmp.pdf";
+	string safetmp = shell_protect(tmpfile);
+	string cmd = "ps2pdf " + shell_protect(file) + " " + safetmp;
+	try {
+	    (void)stdout_to_string(cmd);
+	    cmd = "pdftotext -enc UTF-8 " + safetmp + " -";
+	    dump = stdout_to_string(cmd);
+	} catch (ReadError) {
+	    cout << "\"" << cmd << "\" failed - skipping" << endl;
+	    unlink(tmpfile.c_str());
+	    return;
+	} catch (...) {
+	    unlink(tmpfile.c_str());
+	    throw;
+	}
+	try {
+	    get_pdf_metainfo(safetmp, title, keywords);
+	} catch (...) {
+	    unlink(tmpfile.c_str());
+	    throw;
+	}
+	unlink(tmpfile.c_str());
+    } else if (startswith(mimetype, "application/vnd.sun.xml.") ||
+	       startswith(mimetype, "application/vnd.oasis.opendocument."))
     {
 	// Inspired by http://mjr.towers.org.uk/comp/sxw2text
 	string safefile = shell_protect(file);
 	string cmd = "unzip -p " + safefile + " content.xml";
 	try {
 	    XmlParser xmlparser;
-	    // <?xml version="1.0" encoding="UTF-8"?>
 	    xmlparser.parse_html(stdout_to_string(cmd));
 	    dump = xmlparser.dump;
 	} catch (ReadError) {
@@ -315,7 +341,6 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	cmd = "unzip -p " + safefile + " meta.xml";
 	try {
 	    MetaXmlParser metaxmlparser;
-	    // <?xml version="1.0" encoding="UTF-8"?>
 	    metaxmlparser.parse_html(stdout_to_string(cmd));
 	    title = metaxmlparser.title;
 	    keywords = metaxmlparser.keywords;
@@ -324,7 +349,7 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	    // It's probably best to index the document even if this fails.
 	}
     } else if (mimetype == "application/msword") {
-	string cmd = "antiword " + shell_protect(file);
+	string cmd = "antiword -mUTF-8.txt " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
 	} catch (ReadError) {
@@ -332,8 +357,7 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	    return;
 	}
     } else if (mimetype == "application/vnd.ms-excel") {
-	string cmd = "xls2csv -q0 -d8859-1 " + shell_protect(file);
-	//string cmd = "xls2csv -q0 -dutf-8 " + shell_protect(file);
+	string cmd = "xls2csv -q0 -dutf-8 " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
 	} catch (ReadError) {
@@ -341,8 +365,7 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	    return;
 	}
     } else if (mimetype == "application/vnd.ms-powerpoint") {
-	string cmd = "catppt -d8859-1 " + shell_protect(file);
-	//string cmd = "catppt -dutf-8 " + shell_protect(file);
+	string cmd = "catppt -dutf-8 " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
 	} catch (ReadError) {
@@ -350,6 +373,9 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	    return;
 	}
     } else if (mimetype == "application/vnd.wordperfect") {
+	// Looking at the source of wpd2html and wpd2text I think both output
+	// utf-8, but it's hard to be sure without sample Unicode .wpd files
+	// as they don't seem to be at all well documented.
 	string cmd = "wpd2text " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
@@ -357,20 +383,44 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	    cout << "\"" << cmd << "\" failed - skipping\n";
 	    return;
 	}
-    } else if (mimetype == "text/rtf") {
-	string cmd = "unrtf --nopict --text 2>/dev/null " +
-		     shell_protect(file) +
-		     "|sed '/^### .*/d'";
+    } else if (mimetype == "application/vnd.ms-works") {
+	// wps2text produces UTF-8 output from the sample files I've tested.
+	string cmd = "wps2text " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
 	} catch (ReadError) {
 	    cout << "\"" << cmd << "\" failed - skipping\n";
 	    return;
 	}
+    } else if (mimetype == "text/rtf") {
+	// The --text option unhelpfully converts all non-ASCII characters to
+	// "?" so we use --html instead, which write HTML entities.
+	string cmd = "unrtf --nopict --html 2>/dev/null " + shell_protect(file);
+	MyHtmlParser p;
+	try {
+	    p.parse_html(stdout_to_string(cmd));
+	} catch (ReadError) {
+	    cout << "\"" << cmd << "\" failed - skipping\n";
+	    return;
+	} catch (bool) {
+	    // MyHtmlParser throws a bool to abandon parsing at </body> or when
+	    // indexing is disallowed
+	}
+	if (!p.indexing_allowed) {
+	    cout << "indexing disallowed by meta tag - skipping\n";
+	    return;
+	}
+	dump = p.dump;
+	title = p.title;
+	keywords = p.keywords;
+	sample = p.sample;
     } else if (mimetype == "text/x-perl") {
+	// pod2text's output character set doesn't seem to be documented, but
+	// from inspecting the source it looks like it's probably iso-8859-1.
 	string cmd = "pod2text " + shell_protect(file);
 	try {
 	    dump = stdout_to_string(cmd);
+	    convert_to_utf8(dump, "ISO-8859-1");
 	} catch (ReadError) {
 	    cout << "\"" << cmd << "\" failed - skipping\n";
 	    return;
@@ -389,16 +439,16 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 
     // Produce a sample
     if (sample.empty()) {
-	sample = truncate_to_word(dump, 300);
+	sample = generate_sample(dump, SAMPLE_SIZE);
     } else {
-	sample = truncate_to_word(sample, 300);
+	sample = generate_sample(sample, SAMPLE_SIZE);
     }
 
     // Put the data in the document
     Xapian::Document newdocument;
     string record = "url=" + baseurl + url + "\nsample=" + sample;
     if (!title.empty()) {
-	record += "\ncaption=" + truncate_to_word(title, 100);
+	record += "\ncaption=" + generate_sample(title, TITLE_SIZE);
     }
     record += "\ntype=" + mimetype;
     if (last_mod != (time_t)-1)
@@ -407,11 +457,19 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
 	record += "\nsize=" + long_to_string(size);
     newdocument.set_data(record);
 
-    // Add postings for terms to the document
-    Xapian::termpos pos = 1;
-    pos = index_text(title, newdocument, stemmer, pos);
-    pos = index_text(dump, newdocument, stemmer, pos + 100);
-    pos = index_text(keywords, newdocument, stemmer, pos + 100);
+    // Index the title, document text, and keywords.
+    indexer.set_document(newdocument);
+    if (!title.empty()) {
+	indexer.index_text(title, 2);
+	indexer.increase_termpos(100);
+    }
+    if (!dump.empty()) {
+	indexer.index_text(dump);
+    }
+    if (!keywords.empty()) {
+	indexer.increase_termpos(100);
+	indexer.index_text(keywords);
+    }
 
     newdocument.add_term("T" + mimetype); // mimeType
     string::size_type j;
@@ -436,12 +494,6 @@ index_file(const string &url, const string &mimetype, time_t last_mod, off_t siz
     struct tm *tm = localtime(&last_mod);
     string date_term = "D" + date_to_string(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
     newdocument.add_term(date_term); // Date (YYYYMMDD)
-#if 0 // "Weak" terms aren't currently used by omega
-    date_term.resize(8);
-    date_term[0] = 'W';
-    if (date_term[7] == '3') date_term[7] = '2';
-    newdocument.add_term(date_term); // "Weak" - 10ish day interval
-#endif
     date_term.resize(7);
     date_term[0] = 'M';
     newdocument.add_term(date_term); // Month (YYYYMM)
@@ -484,74 +536,70 @@ static void
 index_directory(size_t depth_limit, const string &dir,
 		map<string, string>& mime_map)
 {
-    struct dirent *ent;
     string path = root + indexroot + dir;
 
     cout << "[Entering directory " << dir << "]" << endl;
 
-    DIR *d = opendir(path.c_str());
-    if (d == NULL) {
-	cout << "Can't open directory \"" << path << "\" - skipping\n";
+    DirectoryIterator d(follow_symlinks);
+    try {
+	d.start(path);
+    } catch (const std::string & error) {
+	cout << error << " - skipping" << endl;
 	return;
     }
-    while ((ent = readdir(d)) != NULL) {
-	struct stat statbuf;
-	// ".", "..", and other hidden files
-	if (ent->d_name[0] == '.') continue;
+    while (d.next()) try {
 	string url = dir;
 	if (!url.empty() && url[url.size() - 1] != '/') url += '/';
-	url += ent->d_name;
+	url += d.leafname();
 	string file = root + indexroot + url;
-#ifdef HAVE_LSTAT
-	if (follow_symlinks) {
-#endif
-	    if (stat(file.c_str(), &statbuf) == -1) {
-		cout << "Can't stat \"" << file << "\" - skipping\n";
-		continue;
-	    }
-#ifdef HAVE_LSTAT
-	} else {
-	    if (lstat(file.c_str(), &statbuf) == -1) {
-		cout << "Can't stat \"" << file << "\" - skipping\n";
-		continue;
-	    }
-	}
-#endif
-	if (S_ISDIR(statbuf.st_mode)) {
-	    if (depth_limit == 1) continue;
-	    try {
-		size_t new_limit = depth_limit;
-		if (new_limit) --new_limit;
-		index_directory(new_limit, url, mime_map);
-	    } catch (...) {
-		cout << "Caught unknown exception in index_directory, rethrowing" << endl;
-		throw;
-	    }
-	    continue;
-	}
-	if (S_ISREG(statbuf.st_mode)) {
-	    string ext;
-	    string::size_type dot = url.find_last_of('.');
-	    if (dot != string::npos) ext = url.substr(dot + 1);
-
-	    map<string,string>::iterator mt = mime_map.find(ext);
-	    if (mt != mime_map.end()) {
-		// It's in our MIME map so we know how to index it.
-		const string & mimetype = mt->second;
+	switch (d.get_type()) {
+	    case DirectoryIterator::DIRECTORY:
+		if (depth_limit == 1) continue;
 		try {
-		    index_file(indexroot + url, mimetype, statbuf.st_mtime,
-			       statbuf.st_size);
-		} catch (NoSuchFilter) {
-		    // FIXME: we ought to ignore by mime-type not extension.
-		    cout << "Filter for \"" << mimetype << "\" not installed - ignoring extension \"" << ext << "\"" << endl;
-		    mime_map.erase(mt);
+		    size_t new_limit = depth_limit;
+		    if (new_limit) --new_limit;
+		    index_directory(new_limit, url, mime_map);
+		} catch (...) {
+		    cout << "Caught unknown exception in index_directory, rethrowing" << endl;
+		    throw;
 		}
+		continue;
+	    case DirectoryIterator::REGULAR_FILE: {
+		string ext;
+		string::size_type dot = url.find_last_of('.');
+		if (dot != string::npos) ext = url.substr(dot + 1);
+
+		map<string,string>::iterator mt = mime_map.find(ext);
+		if (mt != mime_map.end()) {
+		    // Only check the file size if we recognise the extension
+		    // to avoid a call to stat()/lstat() for files we can't
+		    // handle when readdir() tells us the file type.
+		    off_t size = d.get_size();
+		    if (size == 0) {
+			cout << "Skipping empty file: \"" << file << "\"" << endl;
+			continue;
+		    }
+
+		    // It's in our MIME map so we know how to index it.
+		    const string & mimetype = mt->second;
+		    try {
+			time_t mtime = d.get_mtime();
+			index_file(indexroot + url, mimetype, mtime, size);
+		    } catch (NoSuchFilter) {
+			// FIXME: we ought to ignore by mime-type not extension.
+			cout << "Filter for \"" << mimetype << "\" not installed - ignoring extension \"" << ext << "\"" << endl;
+			mime_map.erase(mt);
+		    }
+		}
+		continue;
 	    }
-	    continue;
+	    default:
+		cout << "Not a regular file \"" << file << "\" - skipping\n";
 	}
-	cout << "Not a regular file \"" << file << "\" - skipping\n";
+    } catch (const std::string & error) {
+	cout << error << " - skipping" << endl;
+	continue;
     }
-    closedir(d);
 }
 
 int
@@ -631,6 +679,8 @@ main(int argc, char **argv)
     mime_map["doc"] = "application/msword";
     mime_map["dot"] = "application/msword"; // Word template
     mime_map["wpd"] = "application/vnd.wordperfect";
+    mime_map["wps"] = "application/vnd.ms-works";
+    mime_map["wpt"] = "application/vnd.ms-works"; // Works template
     mime_map["rtf"] = "text/rtf";
     // Other MS formats:
     mime_map["xls"] = "application/vnd.ms-excel";
@@ -647,18 +697,18 @@ main(int argc, char **argv)
 	switch (getopt_ret) {
 	case 'h': {
 	    cout << PROG_NAME" - "PROG_DESC"\n\n"
-"Usage: "PROG_NAME" [OPTIONS] --db DATABASE --url BASEURL [BASEDIR] DIRECTORY\n\n"
+"Usage: "PROG_NAME" [OPTIONS] --db DATABASE [BASEDIR] DIRECTORY\n\n"
 "Options:\n"
 "  -d, --duplicates         set duplicate handling ('ignore' or 'replace')\n"
 "  -p, --preserve-nonduplicates  don't delete unupdated documents in\n"
 "                                duplicate replace mode\n"
 "  -D, --db                 path to database to use\n"
-"  -U, --url                base url DIRECTORY represents\n"
+"  -U, --url                base url DIRECTORY represents (default: /)\n"
 "  -M, --mime-type          additional MIME mapping ext:type\n"
 "  -l, --depth-limit=LIMIT  set recursion limit (0 = unlimited)\n"
 "  -f, --follow             follow symbolic links\n"
 "      --overwrite          create the database anew (the default is to update\n"
-"                           the database already exists)" << endl;
+"                           if the database already exists)" << endl;
 	    print_stemmer_help("     ");
 	    print_help_and_version_help("     ");
 	    return 0;
@@ -743,16 +793,15 @@ main(int argc, char **argv)
 	return 1;
     }
     if (baseurl.empty()) {
-	cerr << PROG_NAME": you must specify a base URL with --url.\n";
-	return 1;
+	cerr << PROG_NAME": --url not specified, assuming `/'.\n";
     }
     // baseurl mustn't end '/' or you end up with the wrong URL
     // (//thing is different to /thing). We could probably make this
     // safe a different way, by ensuring that we don't put a leading '/'
     // on leafnames when scanning a directory, but this will do.
-    if (baseurl[baseurl.length() - 1] == '/') {
+    if (!baseurl.empty() && baseurl[baseurl.length() - 1] == '/') {
 	cout << "baseurl has trailing '/' ... removing ... " << endl;
-	baseurl = baseurl.substr(0, baseurl.length()-1);
+	baseurl.resize(baseurl.size() - 1);
     }
 
     if (optind >= argc || optind + 2 < argc) {
@@ -772,6 +821,7 @@ main(int argc, char **argv)
 	indexroot = ""; // index the whole of root
     }
 
+    int exitcode = 1;
     try {
 	if (!overwrite) {
 	    db = Xapian::WritableDatabase(dbpath, Xapian::DB_CREATE_OR_OPEN);
@@ -782,6 +832,9 @@ main(int argc, char **argv)
 	} else {
 	    db = Xapian::WritableDatabase(dbpath, Xapian::DB_CREATE_OR_OVERWRITE);
 	}
+
+	indexer.set_stemmer(stemmer);
+
 	index_directory(depth_limit, "/", mime_map);
 	if (!skip_duplicates && !preserve_unupdated) {
 	    for (Xapian::docid did = 1; did < updated.size(); ++did) {
@@ -796,17 +849,19 @@ main(int argc, char **argv)
 	}
 	db.flush();
 	// cout << "\n\nNow we have " << db.get_doccount() << " documents.\n";
+	exitcode = 0;
     } catch (const Xapian::Error &e) {
 	cout << "Exception: " << e.get_msg() << endl;
-	return 1;
     } catch (const string &s) {
 	cout << "Exception: " << s << endl;
-	return 1;
     } catch (const char *s) {
 	cout << "Exception: " << s << endl;
-	return 1;
     } catch (...) {
 	cout << "Caught unknown exception" << endl;
-	return 1;
     }
+
+    // If we created a temporary directory then delete it.
+    if (!tmpdir.empty()) rmdir(tmpdir.c_str());
+
+    return exitcode;
 }
