@@ -4,7 +4,7 @@
  * Copyright 2001 James Aylett
  * Copyright 2001,2002 Ananova Ltd
  * Copyright 2002 Intercede 1749 Ltd
- * Copyright 2002,2003,2004,2005,2006 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -32,22 +32,20 @@
 
 #include <assert.h>
 #include <ctype.h>
-#include <errno.h>
+#include "safeerrno.h"
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
-#ifdef HAVE_STRINGS_H
-# include <strings.h> // Needed on Solaris for strcasecmp
-#endif
+#include "strcasecmp.h"
 #include <time.h>
 
-#include <unistd.h>
+#include "safeunistd.h"
 #include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
+#include "safesysstat.h"
+#include "safefcntl.h"
 
 #ifdef HAVE_GETTIMEOFDAY
 #include <sys/time.h>
-#include <unistd.h>
 #elif defined HAVE_FTIME
 #include <sys/timeb.h>
 #endif
@@ -59,22 +57,23 @@
 #endif
 
 #include "date.h"
+#include "datematchdecider.h"
 #include "utils.h"
 #include "omega.h"
 #include "query.h"
 #include "cgiparam.h"
-#include "indextext.h"
 #include "loadfile.h"
+#include "stringutils.h"
 #include "values.h"
 
-#include <xapian/queryparser.h>
-
-#ifdef _MSC_VER
-# define strcasecmp stricmp
-#endif
+#include <xapian.h>
 
 using namespace std;
 
+using Xapian::Utf8Iterator;
+
+using Xapian::Unicode::is_wordchar;
+
 #ifndef SNPRINTF
 #include <stdarg.h>
 
@@ -116,13 +115,16 @@ static string eval_file(const string &fmtfile);
 
 static set<string> termset;
 
+// Holds mapping from term prefix to user prefix (e.g. 'S' -> 'subject:').
+static map<string, string> termprefix_to_userprefix;
+
 static string queryterms;
 
 static string error_msg;
 
 static long int sec = 0, usec = -1;
 
-static const char * DEFAULT_LOG_ENTRY =
+static const char DEFAULT_LOG_ENTRY[] =
 	"$or{$env{REMOTE_HOST},$env{REMOTE_ADDR},-}\t"
 	"[$date{$now,%d/%b/%Y:%H:%M:%S} +0000]\t"
 	"$if{$cgi{X},add,$if{$cgi{MORELIKE},morelike,query}}\t"
@@ -157,12 +159,34 @@ class MyStopper : public Xapian::Stopper {
 		return (t == "was" || t == "what" || t == "when" ||
 			t == "where" || t == "which" || t == "who" ||
 			t == "why" || t == "will" || t == "with");
+	    case 'y':
+		return (t == "you" || t == "your");
 	    default:
 		return false;
 	}
     }
 };
 
+static size_t
+prefix_from_term(string &prefix, const string &term)
+{
+    if (term.empty()) {
+	prefix.resize(0);
+	return 0;
+    }
+    if (term[0] == 'X') {
+	const string::const_iterator begin = term.begin();
+	string::const_iterator i = begin + 1;
+	while (i != term.end() && isupper(static_cast<unsigned char>(*i))) ++i;
+	prefix.assign(begin, i);
+	if (i != term.end() && *i == ':') ++i;
+	return i - begin;
+    }
+
+    prefix = term[0];
+    return 1;
+}
+
 // Don't allow ".." in format names, log file names, etc as this would allow
 // people to open a format "../../etc/passwd" or similar.
 // FIXME: make this check more exact ("foo..bar" is safe)
@@ -195,19 +219,27 @@ set_probabilistic(const string &oldp)
     qp.set_stopper(new MyStopper());
     qp.set_default_op(default_op);
     qp.set_database(db);
+    // std::map::insert() won't overwrite an existing entry, so we'll prefer
+    // probabilistic prefixes to boolean ones in the unlikely event that both
+    // exist for the same term prefix.  We'll also prefer the first specified
+    // prefix.
     map<string, string>::const_iterator pfx = option.lower_bound("prefix,");
-    for (; pfx != option.end() && pfx->first.substr(0, 7) == "prefix,"; ++pfx) {
-	qp.add_prefix(pfx->first.substr(7), pfx->second);
+    for (; pfx != option.end() && startswith(pfx->first, "prefix,"); ++pfx) {
+	string user_prefix = pfx->first.substr(7);
+	qp.add_prefix(user_prefix, pfx->second);
+	termprefix_to_userprefix.insert(make_pair(pfx->second, user_prefix));
     }
     pfx = option.lower_bound("boolprefix,");
-    for (; pfx != option.end() && pfx->first.substr(0, 11) == "boolprefix,"; ++pfx) {
-	qp.add_boolean_prefix(pfx->first.substr(11), pfx->second);
+    for (; pfx != option.end() && startswith(pfx->first, "boolprefix,"); ++pfx) {
+	string user_prefix = pfx->first.substr(11);
+	qp.add_boolean_prefix(user_prefix, pfx->second);
+	termprefix_to_userprefix.insert(make_pair(pfx->second, user_prefix));
     }
 
     try {
 	query = qp.parse_query(query_string);
-    } catch (const char *s) {
-	error_msg = s;
+    } catch (Xapian::QueryParserError &e) {
+	error_msg = e.get_msg();
 	return BAD_QUERY;
     }
 
@@ -225,9 +257,10 @@ set_probabilistic(const string &oldp)
     // Check new query against the previous one
     if (oldp.empty()) return query_string.empty() ? SAME_QUERY : NEW_QUERY;
 
-    // We used to use "word1#word2#" (with trailing #) but some broken old
-    // browsers (versions of MSIE) don't quote # in form GET submissions
-    // and everything after the # gets interpreted as an anchor.
+    // Long, long ago we used "word1#word2#" (with trailing #) but some broken
+    // old browsers (versions of MSIE) don't quote # in form GET submissions
+    // and everything after the # gets interpreted as an anchor.  We now allow
+    // terms like `c#' so we want to avoid '#' anyway.
     //
     // So we switched to using "word1.word2." but that doesn't work if
     // the terms contain "." themselves (e.g. Tapplication/vnd.ms-excel)
@@ -273,16 +306,8 @@ typedef multimap<string, string>::const_iterator FMCI;
 
 void add_bterm(const string &term) {
     string prefix;
-    if (term[0] == 'X') {
-	string::const_iterator i = term.begin() + 1;
-	while (i != term.end() && isupper(static_cast<unsigned char>(*i))) ++i;
-	if (i != term.end() && *i == ':') ++i;
-	prefix = string(term.begin(), i);
-    } else {
-	prefix = term[0];
-    }
-
-    filter_map.insert(multimap<string, string>::value_type(prefix, term));
+    if (prefix_from_term(prefix, term) > 0)
+	filter_map.insert(multimap<string, string>::value_type(prefix, term));
 }
 
 static void
@@ -329,116 +354,118 @@ run_query()
 	}
     }
 
+    Xapian::MatchDecider * mdecider = NULL;
     if (!date_start.empty() || !date_end.empty() || !date_span.empty()) {
-	Xapian::Query date_filter(Xapian::Query::OP_OR,
-				  date_range_filter(date_start, date_end,
-						    date_span),
-				  Xapian::Query("Dlatest"));
-
-	// If no probabilistic query is provided then promote the daterange
-	// filter to be THE query instead of filtering an empty query.
-	if (query.empty()) {
-	    query = date_filter;
+	MCI i = cgi_params.find("DATEVALUE");
+	if (i != cgi_params.end()) {
+	    Xapian::valueno datevalue = string_to_int(i->second);
+	    mdecider = new DateMatchDecider(datevalue, date_start, date_end, date_span);
 	} else {
-	    query = Xapian::Query(Xapian::Query::OP_FILTER, query, date_filter);
+	    Xapian::Query date_filter(Xapian::Query::OP_OR,
+				      date_range_filter(date_start, date_end,
+							date_span),
+				      Xapian::Query("Dlatest"));
+
+	    // If no probabilistic query is provided then promote the daterange
+	    // filter to be THE query instead of filtering an empty query.
+	    if (query.empty()) {
+		query = date_filter;
+	    } else {
+		query = Xapian::Query(Xapian::Query::OP_FILTER, query, date_filter);
+	    }
 	}
     }
 
-    if (enquire && error_msg.empty()) {
-	enquire->set_cutoff(threshold);
+    if (!enquire || !error_msg.empty()) return;
 
-	// Temporary bodge to allow experimentation with Xapian::BiasFunctor
-	MCI i;
-	i = cgi_params.find("bias_weight");
-	if (i != cgi_params.end()) {
-	    Xapian::weight bias_weight = atof(i->second.c_str());
-	    int half_life = 2 * 24 * 60 * 60; // 2 days
-	    i = cgi_params.find("bias_halflife");
-	    if (i != cgi_params.end()) {
-		half_life = atoi(i->second.c_str());
-	    }
-	    enquire->set_bias(bias_weight, half_life);
-	}
-	if (sort_key != Xapian::valueno(-1)) {
+    enquire->set_cutoff(threshold);
+
+    if (sort_key != Xapian::BAD_VALUENO) {
+	if (sort_after) {
+	    enquire->set_sort_by_relevance_then_value(sort_key, sort_ascending);
+	} else {
 	    enquire->set_sort_by_value_then_relevance(sort_key, sort_ascending);
 	}
-	if (collapse) {
-	    enquire->set_collapse_key(collapse_key);
-	}
+    }
+
+    enquire->set_docid_order(docid_order);
+
+    if (collapse) {
+	enquire->set_collapse_key(collapse_key);
+    }
 
 #ifdef HAVE_GETTIMEOFDAY
-	struct timeval tv;
-	if (gettimeofday(&tv, 0) == 0) {
-	    sec = tv.tv_sec;
-	    usec = tv.tv_usec;
-	}
+    struct timeval tv;
+    if (gettimeofday(&tv, 0) == 0) {
+	sec = tv.tv_sec;
+	usec = tv.tv_usec;
+    }
 #elif defined(FTIME_RETURNS_VOID)
-	struct timeb tp;
-	ftime(&tp);
+    struct timeb tp;
+    ftime(&tp);
+    sec = tp.time;
+    usec = tp.millitm * 1000;
+#elif defined(HAVE_FTIME)
+    struct timeb tp;
+    if (ftime(&tp) == 0) {
 	sec = tp.time;
 	usec = tp.millitm * 1000;
-#elif defined(HAVE_FTIME)
-	struct timeb tp;
-	if (ftime(&tp) == 0) {
-	    sec = tp.time;
-	    usec = tp.millitm * 1000;
-	}
+    }
 #else
-	sec = time(NULL);
-	if (sec != (time_t)-1) usec = 0;
+    sec = time(NULL);
+    if (sec != (time_t)-1) usec = 0;
 #endif
-	if (!query.empty()) {
-	    enquire->set_query(query);
-	    // We could use the value of topdoc as first parameter, but we
-	    // need to know the first few items on the mset to fake a
-	    // relevance set for topterms.
-	    //
-	    // If min_hits isn't set, check at least one extra result so we
-	    // know if we've reached the end of the matches or not - then we
-	    // can avoid offering a "next" button which leads to an empty page.
-	    mset = enquire->get_mset(0, topdoc + hits_per_page,
-				     topdoc + max(hits_per_page + 1, min_hits),
-				     &rset);
-	}
-	if (usec != -1) {
+    if (!query.empty()) {
+	enquire->set_query(query);
+	// We could use the value of topdoc as first parameter, but we
+	// need to know the first few items on the mset to fake a
+	// relevance set for topterms.
+	//
+	// If min_hits isn't set, check at least one extra result so we
+	// know if we've reached the end of the matches or not - then we
+	// can avoid offering a "next" button which leads to an empty page.
+	mset = enquire->get_mset(0, topdoc + hits_per_page,
+				 topdoc + max(hits_per_page + 1, min_hits),
+				 &rset, mdecider);
+    }
+    if (usec != -1) {
 #ifdef HAVE_GETTIMEOFDAY
-	    if (gettimeofday(&tv, 0) == 0) {
-		sec = tv.tv_sec - sec;
-		usec = tv.tv_usec - usec;
-		if (usec < 0) {
-		    --sec;
-		    usec += 1000000;
-		}
-	    } else {
-		usec = -1;
+	if (gettimeofday(&tv, 0) == 0) {
+	    sec = tv.tv_sec - sec;
+	    usec = tv.tv_usec - usec;
+	    if (usec < 0) {
+		--sec;
+		usec += 1000000;
 	    }
+	} else {
+	    usec = -1;
+	}
 #elif defined(FTIME_RETURNS_VOID)
-	    ftime(&tp);
+	ftime(&tp);
+	sec = tp.time - sec;
+	usec = tp.millitm * 1000 - usec;
+	if (usec < 0) {
+	    --sec;
+	    usec += 1000000;
+	}
+#elif defined(HAVE_FTIME)
+	if (ftime(&tp) == 0) {
 	    sec = tp.time - sec;
 	    usec = tp.millitm * 1000 - usec;
 	    if (usec < 0) {
 		--sec;
 		usec += 1000000;
 	    }
-#elif defined(HAVE_FTIME)
-	    if (ftime(&tp) == 0) {
-		sec = tp.time - sec;
-		usec = tp.millitm * 1000 - usec;
-		if (usec < 0) {
-		    --sec;
-		    usec += 1000000;
-		}
-	    } else {
-		usec = -1;
-	    }
-#else
-	    usec = time(NULL);
-	    if (usec != -1) {
-		sec = sec - usec;
-		usec = 0;
-	    }
-#endif
+	} else {
+	    usec = -1;
 	}
+#else
+	usec = time(NULL);
+	if (usec != -1) {
+	    sec = sec - usec;
+	    usec = 0;
+	}
+#endif
     }
 }
 
@@ -560,59 +587,54 @@ html_highlight(const string &s, const string &list,
 
     string res;
 
-    AccentNormalisingItor j(s.begin());
-    const AccentNormalisingItor s_end(s.end());
+    Utf8Iterator j(s);
+    const Utf8Iterator s_end;
     while (true) {
-	AccentNormalisingItor first = j;
-	while (first != s_end && !isalnum(static_cast<unsigned char>(*first)))
-	    ++first;
+	Utf8Iterator first = j;
+	while (first != s_end && !is_wordchar(*first)) ++first;
 	if (first == s_end) break;
-	AccentNormalisingItor term_end;
+	Utf8Iterator term_end;
 	string term;
 	string word;
-	string::const_iterator l = j.raw();
-	if (isupper(*first)) {
+	const char *l = j.raw();
+	if (*first < 128 && isupper(*first)) {
 	    j = first;
-	    term = *j;
-	    while (++j != s_end && *j == '.' && ++j != s_end &&
-		   isupper(static_cast<unsigned char>(*j))) {
-		term += *j;
+	    Xapian::Unicode::append_utf8(term, *j);
+	    while (++j != s_end && *j == '.' && ++j != s_end && *j < 128 && isupper(*j)) {
+		Xapian::Unicode::append_utf8(term, *j);
 	    }
-	    if (term.length() < 2 ||
-		(j != s_end && isalnum(static_cast<unsigned char>(*j)))) {
+	    if (term.length() < 2 || (j != s_end && is_wordchar(*j))) {
 		term = "";
 	    }
 	    term_end = j;
 	}
 	if (term.empty()) {
 	    j = first;
-	    while (isalnum(static_cast<unsigned char>(*j))) {
-		term += *j;
+	    while (is_wordchar(*j)) {
+		Xapian::Unicode::append_utf8(term, *j);
 		++j;
 		if (j == s_end) break;
-		if (*j == '&') {
-		    AccentNormalisingItor next = j;
+		if (*j == '&' || *j == '\'') {
+		    Utf8Iterator next = j;
 		    ++next;
-		    if (next == s_end ||
-			!isalnum(static_cast<unsigned char>(*next)))
-			break;
-		    term += '&';
+		    if (next == s_end || !is_wordchar(*next)) break;
+		    term += *j;
 		    j = next;
 		}
 	    }
 	    term_end = j;
-	    if (j != s_end && (*j == '#' || p_plusminus(*j))) {
+	    if (j != s_end && (*j == '+' || *j == '-' || *j == '#')) {
 		string::size_type len = term.length();
 		if (*j == '#') {
 		    term += '#';
 		    do { ++j; } while (j != s_end && *j == '#');
 		} else {
-		    while (j != s_end && p_plusminus(*j)) {
-			term += *j;
+		    while (j != s_end && (*j == '+' || *j == '-')) {
+			Xapian::Unicode::append_utf8(term, *j);
 			++j;
 		    }
 		}
-		if (j != s_end && isalnum(static_cast<unsigned char>(*j))) {
+		if (term.size() - len > 3 || (j != s_end && is_wordchar(*j))) {
 		    term.resize(len);
 		} else {
 		    term_end = j;
@@ -620,19 +642,15 @@ html_highlight(const string &s, const string &list,
 	    }
 	}
 	j = term_end;
-	unsigned char first_char = term[0];
-	lowercase_term(term);
-	int match = -1;
-	// As of 0.8.0, raw terms won't start with a digit.  But we may
-	// be searching an older database where it does, so keep the
-	// isdigit check for now...
-	if (isupper(first_char) || isdigit(first_char)) {
-	    match = word_in_list('R' + term, list);
+	term = Xapian::Unicode::tolower(term);
+	int match = word_in_list(term, list);
+	if (match == -1) {
+	    string stem = "Z";
+	    stem += (*stemmer)(term);
+	    match = word_in_list(stem, list);
 	}
-	if (match == -1)
-	    match = word_in_list(stemmer->stem_word(term), list);
 	if (match >= 0) {
-	    res += html_escape(s.substr(l - s.begin(), first.raw() - l));
+	    res += html_escape(string(l, first.raw() - l));
 	    if (!bra.empty()) {
 		res += bra;
 	    } else {
@@ -644,7 +662,7 @@ html_highlight(const string &s, const string &list,
 		res += colours[match % (sizeof(colours) / sizeof(colours[0]))];
 		res += "\">";
 	    }
-	    word = s.substr(first.raw() - s.begin(), j.raw() - first.raw());
+	    word = string(first.raw(), j.raw() - first.raw());
 	    res += html_escape(word);
 	    if (!bra.empty()) {
 		res += ket;
@@ -652,10 +670,10 @@ html_highlight(const string &s, const string &list,
 		res += "</b>";
 	    }
 	} else {
-	    res += html_escape(s.substr(l - s.begin(), j.raw() - l));
+	    res += html_escape(string(l, j.raw() - l));
 	}
     }
-    if (j != s_end) res += html_escape(s.substr(j.raw() - s.begin()));
+    if (j != s_end) res += html_escape(string(j.raw(), j.left()));
     return res;
 }
 
@@ -691,6 +709,7 @@ static map<string, string> field;
 static Xapian::docid q0;
 static Xapian::doccount hit_no;
 static int percent;
+static Xapian::weight weight;
 static Xapian::doccount collapsed;
 
 static string print_caption(const string &fmt, const vector<string> &param);
@@ -719,7 +738,6 @@ CMD_filterterms,
 CMD_find,
 CMD_fmt,
 CMD_freq,
-CMD_freqs,
 CMD_ge,
 CMD_gt,
 CMD_highlight,
@@ -787,6 +805,7 @@ CMD_unstem,
 CMD_url,
 CMD_value,
 CMD_version,
+CMD_weight,
 CMD_MACRO // special tag for macro evaluation
 };
 
@@ -836,8 +855,6 @@ T(filterterms,	   1, 1, N, 0), // list of terms with a given prefix
 T(find,		   2, 2, N, 0), // find entry in list
 T(fmt,		   0, 0, N, 0), // name of current format
 T(freq,		   1, 1, N, 0), // frequency of a term
-T(freqs,	   0, 0, N, M), // return HTML string listing query terms and
-				// frequencies
 T(ge,		   2, 2, N, 0), // test >=
 T(gt,		   2, 2, N, 0), // test >
 T(highlight,	   2, 4, N, 0), // html escape and highlight words from list
@@ -909,6 +926,7 @@ T(unstem,	   1, 1, N, Q), // return list of probabilistic terms from
 T(url,		   1, 1, N, 0), // url encode argument
 T(value,	   1, 2, N, 0), // return document value
 T(version,	   0, 0, N, 0), // omega version string
+T(weight,	   0, 0, N, 0), // weight of the current hit
 { NULL,{0,	   0, 0, 0, 0}}
 };
 
@@ -1152,20 +1170,42 @@ eval(const string &fmt, const vector<string> &param)
 		value = field[args[0]];
 		break;
 	    case CMD_filesize: {
-		// FIXME: rounding?
-		// FIXME: for smaller sizes give decimal fractions, e.g. "1.4K"
+		// FIXME: rounding?  i18n?
 		int size = string_to_int(args[0]);
-		char buf[200] = "";
-		if (size && size < 1024) {
-		    my_snprintf(buf, sizeof(buf), "%d bytes", size);
+		int intpart = size;
+		int fraction = -1;
+		const char * format = 0;
+		if (size < 0) {
+		    // Negative size -> empty result.
+		} else if (size == 1) {
+		    format = "%d byte";
+		} else if (size < 1024) {
+		    format = "%d bytes";
 		} else if (size < 1024*1024) {
-		    my_snprintf(buf, sizeof(buf), "%dK", int(size/1024));
+		    intpart = size / 1024;
+		    fraction = size % 1024;
+		    format = "%d.%cK";
 		} else if (size < 1024*1024*1024) {
-		    my_snprintf(buf, sizeof(buf), "%dM", int(size/1024/1024));
+		    intpart = size / (1024 * 1024);
+		    fraction = size % (1024 * 1024);
+		    format = "%d.%cM";
 		} else {
-		    my_snprintf(buf, sizeof(buf), "%dG", int(size/1024/1024/1024));
+		    intpart = size / (1024 * 1024 * 1024);
+		    fraction = size % (1024 * 1024 * 1024);
+		    format = "%d.%cG";
+		}
+		if (format) {
+		    char buf[200];
+		    int len;
+		    if (fraction == -1) {
+			len = my_snprintf(buf, sizeof(buf), format, intpart);
+		    } else {
+			fraction = (fraction * 10 / 1024) + '0';
+			len = my_snprintf(buf, sizeof(buf), format, intpart, fraction);
+		    }
+		    if (len < 0 || (unsigned)len > sizeof(buf)) len = sizeof(buf);
+		    value.assign(buf, len);
 		}
-		value = buf;
 		break;
 	    }
 	    case CMD_filters:
@@ -1176,7 +1216,7 @@ eval(const string &fmt, const vector<string> &param)
 		term.skip_to(args[0]);
 		while (term != db.allterms_end()) {
 		    string t = *term;
-		    if (t.substr(0, args[0].size()) != args[0]) break;
+		    if (!startswith(t, args[0])) break;
 		    value = value + t + '\t';
 		    ++term;
 		}
@@ -1212,11 +1252,6 @@ eval(const string &fmt, const vector<string> &param)
 		    value = int_to_string(db.get_termfreq(args[0]));
 		}
 		break;
-	    case CMD_freqs:
-		// for backward compatibility
-		value = eval("$map{$queryterms,$_:&nbsp;$nice{$freq{$_}}}",
-			     param);
-		break;
             case CMD_ge:
 		if (string_to_int(args[0]) >= string_to_int(args[1]))
 		    value = "true";
@@ -1292,10 +1327,10 @@ eval(const string &fmt, const vector<string> &param)
 		value = value.substr(i, value.find('/', i) - i);
 		// remove user@ or user:password@
 		i = value.find('@');
-		if (i != string::npos) value = value.substr(i + 1);
+		if (i != string::npos) value.erase(0, i + 1);
 		// remove :port
 		i = value.find(':');
-		if (i != string::npos) value = value.substr(0, i);
+		if (i != string::npos) value.resize(i);
 		break;
 	    }
 	    case CMD_html:
@@ -1507,6 +1542,9 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_now: {
 		char buf[64];
 		my_snprintf(buf, sizeof(buf), "%lu", (unsigned long)time(NULL));
+		// MSVC's snprintf omits the zero byte if the string if
+		// sizeof(buf) long.
+		buf[sizeof(buf) - 1] = '\0';
 		value = buf;
 		break;
 	    }
@@ -1677,13 +1715,24 @@ eval(const string &fmt, const vector<string> &param)
 		    if (static_cast<size_t>(-start) >= args[0].size()) {
 			start = 0;
 		    } else {
-			start = static_cast<int>(args[0].size()) - start;
+			start = static_cast<int>(args[0].size()) + start;
 		    }
+		} else {
+		    if (static_cast<size_t>(start) >= args[0].size()) break;
 		}
 		size_t len = string::npos;
 		if (args.size() > 2) {
 		    int int_len = string_to_int(args[2]);
-		    if (int_len >= 0) len = size_t(int_len);
+		    if (int_len >= 0) {
+			len = size_t(int_len);
+		    } else {
+			len = args[0].size() - start;
+			if (static_cast<size_t>(-int_len) >= len) {
+			    len = 0;
+			} else {
+			    len -= static_cast<size_t>(-int_len);
+			}
+		    }
 		}
 		value = args[0].substr(start, len);
 		break;
@@ -1710,6 +1759,9 @@ eval(const string &fmt, const vector<string> &param)
 		if (usec != -1) {
 		    char buf[64];
 		    my_snprintf(buf, sizeof(buf), "%ld.%06ld", sec, usec);
+		    // MSVC's snprintf omits the zero byte if the string if
+		    // sizeof(buf) long.
+		    buf[sizeof(buf) - 1] = '\0';
 		    value = buf;
 		}
 		break;
@@ -1725,7 +1777,7 @@ eval(const string &fmt, const vector<string> &param)
 
 		    // List of expand terms
 		    Xapian::ESet eset;
-		    ExpandDeciderOmega decider(db);
+		    OmegaExpandDecider decider(db, &termset);
 
 		    if (!rset.empty()) {
 			eset = enquire->get_eset(howmany * 2, rset, &decider);
@@ -1744,35 +1796,15 @@ eval(const string &fmt, const vector<string> &param)
 			eset = enquire->get_eset(howmany * 2, tmp, &decider);
 		    }
 
+		    // Don't show more than one word with the same stem.
+		    set<string> stems;
 		    Xapian::ESetIterator i;
-		    set<string> seen;
-		    {
-			if (!stemmer) {
-			    stemmer = new Xapian::Stem(option["stemmer"]);
-			}
-			// Exclude terms "similar" to those already in
-			// the query
-			set<string>::const_iterator t;
-			for (t = termset.begin(); t != termset.end(); ++t) {
-			    string term = *t;
-			    if (term[0] == 'R') {
-				term.erase(0, 1);
-				term = stemmer->stem_word(term);
-			    }
-			    seen.insert(term);
-			}
-		    }
-		    MyStopper stopper;
 		    for (i = eset.begin(); i != eset.end(); ++i) {
-			string term = *i;
-			if (term[0] == 'R') {
-			    term.erase(0, 1);
-			    if (stopper(term)) continue;
-			    term = stemmer->stem_word(term);
-			}
-			if (seen.find(term) != seen.end()) continue;
-			seen.insert(term);
-			value += *i;
+			string term(*i);
+			string stem = (*stemmer)(term);
+			if (stems.find(stem) != stems.end()) continue;
+			stems.insert(stem);
+			value += term;
 			value += '\t';
 			if (--howmany == 0) break;
 		    }
@@ -1861,6 +1893,9 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_version:
 		value = "Xapian - "PACKAGE" "VERSION;
 		break;
+	    case CMD_weight:
+		value = double_to_string(weight);
+		break;
 	    default: {
 		args.insert(args.begin(), param[0]);
 		int macro_no = func->second->tag - CMD_MACRO;
@@ -1905,36 +1940,63 @@ eval_file(const string &fmtfile)
 }
 
 extern string
-pretty_term(const string & term)
+pretty_term(string term)
 {
-    if (term.empty()) return term;
+    // Just leave empty strings and single characters alone.
+    if (term.length() <= 1) return term;
 
-    if (term.length() >= 2 && term[0] == 'R') {
-	string result = term.substr(1);
-	result[0] = toupper(static_cast<unsigned char>(result[0]));
-	return result;
+    // Assume unprefixed terms are unstemmed.
+    if (!isupper(term[0])) return term;
+
+    // FIXME: keep this for now in case people are still generating 'R' terms?
+    // But if we assumed unprefixed terms are unstemmed, what use is this?
+    if (term[0] == 'R') {
+	term.erase(0, 1);
+	term[0] = toupper(static_cast<unsigned char>(term[0]));
+	return term;
     }
 
-    // If there's an unstemmed version in the query, use that.
-    // If the multiple forms with the same stem appear, we arbitrarily pick
-    // the first.
-    Xapian::TermIterator i = qp.unstem_begin(term);
-    if (i != qp.unstem_end(term)) return *i;
+    // Handle stemmed terms.
+    bool stemmed = (term[0] == 'Z');
+    if (stemmed) {
+	// First of all, check if a term in the query stemmed to this one.
+	Xapian::TermIterator u = qp.unstem_begin(term);
+	// There might be multiple words with the same stem, but we only want
+	// one so just take the first.
+	if (u != qp.unstem_end(term)) return *u;
 
-    // If the term wasn't indexed unstemmed, it's probably a non-term
-    // e.g. "litr" - the stem of "litre"
-    // FIXME: perhaps ought to check termfreq > some threshold
-    if (!db.term_exists('R' + term))
-	return term + '.';
-
-    if (!stemmer) {
-	stemmer = new Xapian::Stem(option["stemmer"]);
+	// Remove the 'Z'.
+	term.erase(0, 1);
     }
 
-    // The term is present unstemmed, but if it would stem further it still
-    // needs protecting.
-    if (stemmer->stem_word(term) != term)
-	return term + '.';
+    bool add_quotes = false;
+
+    // Check if the term has a prefix.
+    if (isupper(term[0])) {
+	// See if we have this prefix in the termprefix_to_userprefix map.  If
+	// so, just reverse the mapping (e.g. turn 'Sfish' into 'subject:fish').
+	string prefix;
+	size_t prefix_len = prefix_from_term(prefix, term);
+
+	map<string, string>::const_iterator i;
+	i = termprefix_to_userprefix.find(prefix);
+	if (i != termprefix_to_userprefix.end()) {
+	    string user_prefix = i->second;
+	    user_prefix += ':';
+	    term.replace(0, prefix_len, user_prefix);
+	} else {
+	    // We don't have a prefix mapping for this, so just set a flag to
+	    // add quotes around the term.
+	    add_quotes = true;
+	}
+    }
+
+    if (stemmed) term += '.';
+
+    if (add_quotes) {
+	term.insert(0, "\"");
+	term.append("\"");
+    }
 
     return term;
 }
@@ -1944,6 +2006,7 @@ print_caption(const string &fmt, const vector<string> &param)
 {
     q0 = *(mset[hit_no]);
 
+    weight = mset[hit_no].get_weight();
     percent = mset.convert_to_percent(mset[hit_no]);
     collapsed = mset[hit_no].get_collapse_count();
 
@@ -2055,32 +2118,7 @@ ensure_query_parsed()
 		force_first_page = false;
 		val = cgi_params.find("xFILTERS");
 		string xfilters;
-		if (val != cgi_params.end()) {
-		    xfilters = val->second;
-		} else {
-		    // compatibility with older xB/xDATE/... scheme
-		    val = cgi_params.find("xB");
-		    if (val != cgi_params.end())
-			xfilters = val->second + filter_sep;
-		    static const char * check_vars[] = {
-			"DATE1", "DATE2", "DAYSMINUS", NULL
-		    };
-		    for (const char **pv = check_vars; *pv; ++pv) {
-			val = cgi_params.find('x' + string(*pv));
-			if (val != cgi_params.end()) xfilters += val->second;
-			xfilters += filter_sep;
-		    }
-		    val = cgi_params.find("xDEFAULTOP");
-		    if (val == cgi_params.end() && xfilters.length() == 3) {
-			// no x values, so don't force first page
-			xfilters = filters;
-		    } else {
-			char ch = 'O';
-			if (val != cgi_params.end() && val->second == "and")
-			    ch = 'A';
-			xfilters[xfilters.length() - 1] = ch;
-		    }
-		}
+		if (val != cgi_params.end()) xfilters = val->second;
 		if (filters != xfilters) {
 		    // Filters changed since last query
 		    force_first_page = true;
@@ -2118,9 +2156,6 @@ ensure_query_parsed()
 	// this is unhelpful.
 	bool raw_search = false;
 	val = cgi_params.find("RAWSEARCH");
-	// In Omega <= 0.6.3, RAWSEARCH was RAW_SEARCH - renamed to be
-	// consistent with the naming of other CGI parameters.
-	if (val == cgi_params.end()) val = cgi_params.find("RAW_SEARCH");
 	if (val != cgi_params.end()) {
 	    raw_search = bool(atol(val->second.c_str()));
 	}
@@ -2166,3 +2201,75 @@ ensure_match()
 	    last = topdoc + hits_per_page;
     }
 }
+
+// OmegaExpandDecider methods.
+
+OmegaExpandDecider::OmegaExpandDecider(const Xapian::Database & db_,
+				       set<string> * querytermset)
+    : db(db_)
+{
+    // We'll want the stemmer for testing matches anyway.
+    if (!stemmer)
+	stemmer = new Xapian::Stem(option["stemmer"]);
+    if (querytermset) {
+	set<string>::const_iterator i;
+	for (i = querytermset->begin(); i != querytermset->end(); ++i) {
+	    string term(*i);
+	    if (term.empty()) continue;
+
+	    unsigned char ch = term[0];
+	    bool stemmed = (ch == 'Z');
+	    if (stemmed) {
+	       term.erase(0, 1);
+	       if (term.empty()) continue;
+	       ch = term[0];
+	    }
+
+	    if (isupper(ch)) {
+		string prefix;
+		size_t prefix_len = prefix_from_term(prefix, term);
+		term.erase(0, prefix_len);
+	    }
+
+	    if (!stemmed) term = (*stemmer)(term);
+
+	    exclude_stems.insert(term);
+	}
+    }
+}
+
+bool
+OmegaExpandDecider::operator()(const string & term) const
+{
+    unsigned char ch = term[0];
+
+    // Reject terms with a prefix.
+    if (isupper(ch)) return false;
+
+    {
+	MyStopper stopper;
+	// Don't suggest stopwords.
+	if (stopper(term)) return false;
+    }
+
+    // Reject small numbers.
+    if (term.size() < 4 && isdigit(ch)) return false;
+
+    // Reject terms containing a space.
+    if (term.find(' ') != string::npos) return false;
+
+    // Skip terms with stems in the exclude_stems set, to avoid suggesting
+    // terms which are already in the query in some form.
+    string stem = (*stemmer)(term);
+    if (exclude_stems.find(stem) != exclude_stems.end())
+	return false;
+
+    // Ignore terms that only occurs once (hapaxes) since they aren't
+    // useful for finding related documents - they only occurs in a
+    // document that's already been marked as relevant.
+    // FIXME: add an expand option to ignore terms where
+    // termfreq == rtermfreq.
+    if (db.get_termfreq(term) <= 1) return false;
+
+    return true;
+}
