@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001 Sam Liddicott
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -33,15 +33,16 @@
 #include <string>
 #include <vector>
 
-#include <errno.h>
+#include "safeerrno.h"
 #include <stdio.h>
-#include <unistd.h>
+#include <time.h>
+#include "safeunistd.h"
 
 #include "commonhelp.h"
 #include "hashterm.h"
-#include "indextext.h"
 #include "loadfile.h"
 #include "myhtmlparse.h"
+#include "utf8truncate.h"
 #include "utils.h"
 
 #include "gnu_getopt.h"
@@ -51,13 +52,6 @@ using namespace std;
 #define PROG_NAME "scriptindex"
 #define PROG_DESC "index arbitrary data as described by an index script"
 
-#ifdef __WIN32__
-inline unsigned int sleep(unsigned int secs) {
-    _sleep(secs * 1000);
-    return 0;
-}
-#endif
-
 static bool verbose;
 static int addcount;
 static int repcount;
@@ -234,16 +228,6 @@ parse_index_script(const string &filename)
 			    code = Action::TRUNCATE;
 			    arg = YES;
 			    takes_integer_argument = true;
-			    if (!actions.empty() &&
-				actions.back().get_action() == Action::LOAD) {
-				/* Turn "load truncate=n" into "load" with
-				 * num_arg n, so that we don't needlessly
-				 * allocate memory and read data we're just
-				 * going to ignore.
-				 */
-				actions.pop_back();
-				code = Action::TRUNCATE;
-			    }
 			}
 			break;
 		    case 'u':
@@ -327,6 +311,19 @@ parse_index_script(const string &filename)
 			}
 			useless_weight_pos = j - s.begin();
 			break;
+		    case Action::TRUNCATE:
+			if (!actions.empty() &&
+			    actions.back().get_action() == Action::LOAD) {
+			    /* Turn "load truncate=n" into "load" with
+			     * num_arg n, so that we don't needlessly
+			     * allocate memory and read data we're just
+			     * going to ignore.
+			     */
+			    actions.pop_back();
+			    code = Action::LOAD;
+			}
+			actions.push_back(Action(code, val));
+			break;
 		    case Action::UNIQUE:
 			if (boolmap.find(val) == boolmap.end())
 			    boolmap[val] = Action::UNIQUE;
@@ -411,15 +408,15 @@ parse_index_script(const string &filename)
 
 static bool
 index_file(const char *fname, istream &stream,
-	   Xapian::WritableDatabase &database, Xapian::Stem &stemmer)
+	   Xapian::WritableDatabase &database, Xapian::TermGenerator &indexer)
 {
     string line;
     size_t line_no = 0;
     while (!stream.eof() && getline(stream, line)) {
 	++line_no;
 	Xapian::Document doc;
+	indexer.set_document(doc);
 	Xapian::docid docid = 0;
-	Xapian::termpos wordcount = 0;
 	map<string, list<string> > fields;
 	bool seen_content = false;
 	while (!line.empty()) {
@@ -477,16 +474,17 @@ index_file(const char *fname, istream &stream,
 			}
 			break;
 		    case Action::INDEX:
-			wordcount = index_text(value, doc, stemmer,
-					       i->get_num_arg(),
-					       i->get_string_arg(), wordcount);
+			indexer.index_text(value,
+					   i->get_num_arg(),
+					   i->get_string_arg());
 			break;
 		    case Action::INDEXNOPOS:
 			// No positional information so phrase searching
 			// won't work.  However, the database will use much
 			// less diskspace.
-			index_text(value, doc, stemmer, i->get_num_arg(),
-				   i->get_string_arg());
+			indexer.index_text_without_positions(value,
+							     i->get_num_arg(),
+							     i->get_string_arg());
 			break;
 		    case Action::BOOLEAN:
 			doc.add_term(i->get_string_arg() + value);
@@ -500,7 +498,7 @@ index_file(const char *fname, istream &stream,
 			break;
 		    }
 		    case Action::LOWER:
-			lowercase_term(value);
+			value = Xapian::Unicode::tolower(value);
 			break;
 		    case Action::LOAD: {
 			bool truncated = false;
@@ -513,20 +511,9 @@ index_file(const char *fname, istream &stream,
 			if (!truncated) break;
 			/* FALLTHRU (conditionally) */
 		    }
-		    case Action::TRUNCATE: {
-			string::size_type l = i->get_num_arg();
-			if (l < value.size()) {
-			    // Trim back to (and including) previous whitespace.
-			    while (l > 0 && !isspace(static_cast<unsigned char>(value[l - 1]))) --l;
-			    while (l > 0 && isspace(static_cast<unsigned char>(value[l - 1]))) --l;
-
-			    // If the first word is too long, just truncate it.
-			    if (l == 0) l = i->get_num_arg();
-
-			    value = value.substr(0, l);
-			}
+		    case Action::TRUNCATE:
+			utf8_truncate(value, i->get_num_arg());
 			break;
-		    }
 		    case Action::UNHTML: {
 			MyHtmlParser p;
 			try {
@@ -590,12 +577,6 @@ again:
 			if (yyyymmdd.empty()) break;
 			// Date (YYYYMMDD)
 			doc.add_term("D" + yyyymmdd);
-#if 0 // "Weak" terms aren't currently used by omega
-			yyyymmdd.resize(7);
-			if (yyyymmdd[6] == '3') yyyymmdd[6] = '2';
-			// "Weak" - 10ish day interval
-			newdocument.add_term("W" + yyyymmdd);
-#endif
 			yyyymmdd.resize(6);
 			// Month (YYYYMM)
 			doc.add_term("M" + yyyymmdd);
@@ -679,14 +660,13 @@ main(int argc, char **argv)
 	{ "version",	no_argument,	NULL, 'V' },
 	{ "stemmer",	required_argument,	NULL, 's' },
 	{ "overwrite",	no_argument,	NULL, 'o' },
-	{ "quiet",	no_argument,	NULL, 'q' },
 	{ "verbose",	no_argument,	NULL, 'v' },
 	{ 0, 0, NULL, 0 }
     };
 
     bool more = true, show_help = false;
     while (more) {
-	switch (gnu_getopt_long(argc, argv, "uqv", longopts, NULL)) {
+	switch (gnu_getopt_long(argc, argv, "vs:hV", longopts, NULL)) {
 	    case EOF:
 		more = false;
 		break;
@@ -701,12 +681,6 @@ main(int argc, char **argv)
 	    case 'o': // --overwrite
 		database_mode = Xapian::DB_CREATE_OR_OVERWRITE;
 		break;
-	    case 'u':
-		// Update is now the default, so ignore -u for compatibility.
-		break;
-	    case 'q':
-		// --quiet no longer does anything; ignore for compatibility.
-		break;
 	    case 'v':
 		verbose = true;
 		break;
@@ -758,19 +732,22 @@ main(int argc, char **argv)
 	    }
 	}
 
+	Xapian::TermGenerator indexer;
+	indexer.set_stemmer(stemmer);
+
 	addcount = 0;
 	repcount = 0;
 	delcount = 0;
 
 	if (argc == 2) {
 	    // Read from stdin.
-	    index_file("<stdin>", cin, database, stemmer);
+	    index_file("<stdin>", cin, database, indexer);
 	} else {
 	    // Read file(s) listed on the command line.
 	    for (int i = 2; i < argc; ++i) {
 		ifstream stream(argv[i]);
 		if (stream) {
-		    index_file(argv[i], stream, database, stemmer);
+		    index_file(argv[i], stream, database, indexer);
 		} else {
 		    cout << "Can't open file " << argv[i] << endl;
 		}
